{% load i18n %}
{% load evelinks %}

{% if character.location %}
    {% with solar_system=character.location.location.eve_solar_system %}
        <a href="{{ solar_system.name|dotlan_solar_system_url }}" target="_blank">
            {{ solar_system.name }}
        </a>
        <span class="{% if solar_system.is_high_sec %}text-high-sec{% elif solar_system.is_low_sec %}text-low-sec{% else %}text-null-sec{% endif%}">
            {{ solar_system.security_status|floatformat:1 }}
        </span>
        /
        {{ solar_system.eve_constellation.eve_region.name }}
    {% endwith %}
{% else %}
    <span class="text-muted">{% translate "(no data yet)" %}</span>
{% endif %}
