# Generated by Django 4.0.10 on 2023-12-13 23:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("memberaudit", "0016_extend_characterupdatestatus"),
    ]

    operations = [
        migrations.AlterField(
            model_name="characterasset",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
    ]
